# -*- coding: utf-8 -*-
"""
Created on Fri Nov 26 11:15:50 2021
@author: geotw
@author: geomws (edits)
For extracting rainfall data from ISIMIP nc4 files
"""

#Imports all relevant components
from netCDF4 import Dataset
import netCDF4 as nc
import numpy as np
import pandas as pd
import ogr,osr
import shapely
import shapely.geometry
import json
import os
import gdal
import math
import calendar
import time
import re
import argparse
from datetime import date, timedelta, datetime

start_time = time.time()

#[1] NETC_SEARCH
#Takes the directory inputted by the user in L1 of the text file (netc_dir)
#Searches for any .nc4 file in that directory (using os.listdir)
#Returns a 'resfiles' as a list of those filenames joined with the input directory location
def netc_search(netc_dir):
    resfiles = [os.path.join(netc_dir,f) 
    for f in os.listdir(netc_dir) if f.endswith('.nc4')]
    return resfiles

#[2] DAYS_LIST
#Opens all netcdf files in directory, creates new data table to create single list of all dates (daylist)
#Uses netc_inputs (list of .nc4 files) as input
def days_list(netc_inputs): 
    daylist = []
    #MFDataset makes variables spanning multiple files appear as though they were in one file
    nc_open = nc.MFDataset(netc_inputs)
    #time_since takes the units of the 'time' column of the nc4 file
    time_since = nc_open.variables['time'].units
    #match checks for matches at the beginning of a string and returns '1661-1-1' as a string
    match = re.search(r'\d{4}-\d{1}-\d{1}', time_since)
    ##strptime converts a string into a datetime object and returns it in 'start' as a date with padded zeros (.date = idealised date)
    start = datetime.strptime(match.group(), '%Y-%m-%d').date()
    #opens time variable in nc4 files
    days = nc_open.variables['time'][:]
    #works through list
    for i,day in enumerate(days):
        day = int(day)
        #Timedelta returns difference between two datetime instances
        delta = timedelta(day)
        offset = start + delta
        daylist.append(offset)
    return(daylist)

#[3] DAY_FRAME
#Converts days_list into Datetime data type
def day_frame(daylist):
    #sets up three lists for years, months and day numbers and populates them
    yearlist = []
    monthlist = []
    daynumlist = []
    for i,day in enumerate(daylist):
        a = day.year
        b = day.month
        c = day.day
        yearlist.append(a)
        monthlist.append(b)
        daynumlist.append(c)
    #Uses pandas to set up a data frame from daylist
    dayframe = pd.DataFrame(daylist)
    #Converts it to a datetime format
    dayframe = pd.DataFrame(pd.DatetimeIndex(daylist))
    #gives dayframe a column name of date where the data are stored
    dayframe = dayframe.rename(index =str,columns = {0:'date'})
    #Creates a new column called 'frameref' that's  another index (the existing one will change)
    dayframe['frameref'] = range(0, len(dayframe))
    #Now adds columns of years, months and days to dayframe
    dayframe['year'] = yearlist
    dayframe['month'] = monthlist
    dayframe['day'] = daynumlist
    #Assigns datetime to the index (which was stored in the date column)
    dayframe = dayframe.set_index(pd.DatetimeIndex(dayframe['date']))
    return (dayframe)

#[4] FILENAME
    #Reads in first and last file organised by date. Gives us the date range to name outputs
    #Long hand for making auto name - may wish to just make it an input and comment out
    #This needs editing for each code (PR, T, Q)
#def filename(netc_file,netc_file2):
#    name = netc_file.split("\\")[-1]
#    name = name.split(".")[0]
#    name = name.replace("EWEMBI_","")
#    name = name.replace("r1i1p1_","")
#    name2 = netc_file2.split("\\")[-1]
#    name2 = name2.split(".")[0]
#    name2 = name2.split("-")[2]
#    name = name.replace((name.split("-")[-1]),name2)
#    return(name)

#[5] GENERAL CALCULATOR (not just runoff).
    #Open data. 
def runoff_calc(dateselect,nc_open,clip_domain,tablength):
    #Set up output table (outtab) with same rows as tablelength and 1 column
    outtab = np.zeros([tablength,1])
    #Identify the start month from the lowest frameref value in the range (and vv for the end)
    monthstart = min(dateselect.frameref)
    monthend = max(dateselect.frameref)+1 
    #Open the pr variable from the nc4 files for the given month (at all Africa lat/longs - see ranges)
    pr = nc_open.variables['pr'][monthstart:monthend,99:261 ,319:480]
    #pr = nc_open.variables['pr'][:][monthstart:monthend]
    #Convert to mm/day.
    pr = pr * 86400
    #Set up an array of zeros the same size as pr at a single timestep (i.e. a lat/long matrix)
    outframe = np.zeros_like(pr[0])
    #Sets up a frame that takes each day of pr and then adds that to the outframe to sum the monthly total
    for i,frame in enumerate(pr):
        a = frame
        outframe = outframe + a
    #Restructures the array for outr lat/lon list
    count = 0
    for j,element in np.ndenumerate(outframe):
        outtab[count,0] = (element)
        count = count + 1
    return(outtab)

#[6] MONTHLY AVERAGE
#Combines daily calculations into monthly values over time period. Creates csv of average monthly data (rename this)    
def monthly_average(tab,months,tablength,name,filepath):
    #Drop the first two (lat/lon) columns of tab
    tabs = tab[:,2:]
    #Set up month_out variable
    month_out = []
    #create output tab (just the lat/long columns of tab)
    monthtab = tab[:,0:2]
    #For each of 12 months...
    for i,month in enumerate(months):
        #Append a text description of the month to month_out 
        month_out.append(str((calendar.month_name[month])))
        #Extracts all values for the given month for each year
        tabslice = tabs[:, i::12]
        #Takes mean of all these 
        monthmeans = np.mean(tabslice,axis=1)
        #Reshapes the monthmeans 
        monthmeans = np.reshape(monthmeans,(tablength,1))
        #Appends to monthtab
        monthtab = np.append(monthtab,monthmeans,axis=1)
    #Create a pandas dataframe of these data
    outdf = pd.DataFrame(data=monthtab)
    #Drop any rows with all 'NA' data
    outdf = outdf.dropna(how='any')
    #For each column after lat/longs in outdf use the month name in 'monthout' as a label
    outdf.columns = [month_out[x-2] if x > 1 else x for x in outdf.columns]
    #Rename lat and long columns too
    outdf = outdf.rename(columns={0:'Latitude',1:'Longitude'})
    #Export dataframe to a csv file using pre-supplied filepath
    outdf.to_csv(os.path.join(filepath,str(name+"_MonthlyAverageRain_Data.csv")) ,sep=',',index=False)
    print("Monthly Averages Computed")
    return (outdf)

#[7] SPATIAL REFERENCE (used in MONTHLY WRITE)
    #Build TIF output for GIS, gives wkt (well known text)  
def spatialref(sp_rf):
    sr = osr.SpatialReference()
    sr.ImportFromEPSG(sp_rf)
    sr = sr.ExportToWkt()
    return(sr)

#[8] MONTHLY WRITE
    #Creates geotiff of the above and masking it. Puts in output folder  
driver = gdal.GetDriverByName('GTiff')
spatial_reference = 4326
nodata = -9999
def monthly_write(month_table,months,mask_lat,mask_lon,name,filepath):
    ##figure out the dimensions of the data for the gtiff
    sr = spatialref(spatial_reference)
    cellsize = abs(mask_lat[0] - mask_lat[1])
    xmin = min(mask_lon)
    xmax = max(mask_lon)
    ymin = min(mask_lat)
    ymax = max(mask_lat)
    num_cols = len(mask_lon)
    num_rows = len(mask_lat)
    month_sort = month_table
    for i,month in enumerate(months):
        month_name = str((calendar.month_name[month]))
        ##create file names
        tiff_name = os.path.join(filepath,str(name+month_name+"_Average.tif"))
        # get the data, split to array
        month_data = np.array(month_sort[[month_name]])
        month_array = np.empty((num_rows, num_cols), dtype=np.float32)
        month_array[:] = nodata
        for j, (lon, lat) in enumerate(zip(month_sort.Longitude,month_sort.Latitude)):
            x = math.floor((lon - xmin) / cellsize)
            y = math.floor((lat - ymin) / cellsize)
            month_array[y, x] = month_data[j]
        month_array = np.flipud(month_array)
    ##https://gis.stackexchange.com/questions/318050/writing-numpy-arrays-to-irregularly-shaped-multiband-raster
        #month_array = month_data.reshape((len(mask_lat),len(mask_lon)))
        new_dataset = driver.Create(tiff_name,
                                    num_cols,    # number of columns
                                    num_rows,    # number of rows
                                    1,              # number of bands
                                    gdal.GDT_Float32)  # datatype of the raster
        new_dataset.SetProjection(sr)
        new_dataset.SetGeoTransform((xmin,cellsize,0,ymax,0,-cellsize))
        new_band = new_dataset.GetRasterBand(1)
        new_band.SetNoDataValue(-9999)
        new_band.WriteArray(month_array)
        new_band.FlushCache()
        new_dataset = None
    print("Monthly Raster WriteOut Complete")
    return()

#[9] CLIPMASK - optional
#Reads in mask, converts to Json type. Looks through lat/long list and gives a false/true whether inside mask or not    
#def clipmask(clip_domain,lat,lon):
#    tab = np.zeros([(len(lat)*len(lon)),2])
#    count = 0
#    for i in range(0,len(lat)):
#        for j in range(0,len(lon)):
#            tab[count,0] = lat[i]
#            tab[count,1] = lon[j]
#            count = count + 1
#    shp = ogr.Open(clip_domain)
#    lyr = shp.GetLayer(0)
#    feat = lyr.GetFeature(0)
#    first = feat.ExportToJson()
#    first_a = json.loads(first)
#    poly_data = first_a["geometry"]["coordinates"][0]
#    poly = shapely.geometry.Polygon(poly_data)
#    mask = np.array([poly.contains(shapely.geometry.Point(y, x)) for x, y in tab])
#    return(mask)
   


#MAIN FUNCTION

def main():
    #Take file names from imput text file and prompt command line input
    #Create list (params) each containing a line of input and strip any tailing characters
    parser = argparse.ArgumentParser()
    parser.add_argument("--file","-f",type=str,required =True)
    args = parser.parse_args()
    file = args.file
    File = open(file)
    params = []
    for line in File:
        params.append(line)
    netc_dir = params[0].rstrip()
    clip_domain = params[1].rstrip()
    filepath = params[2].rstrip()
    startyear = params[3].rstrip()
    endyear = params[4].rstrip()
    filename = params[5].rstrip()

    ##Create the range of dates (i.e. add day month after the years)
    startdate  = datetime(int(startyear),1,1)
    enddate = datetime(int(endyear),12,31)

    #Call function netc_search (above) to get list of .nc4 files to use (netc_inputs)
    netc_inputs = netc_search(netc_dir)
    #Apply days_list function to file to extract list of dates (daylist)
    daylist = days_list(netc_inputs)
    #Apply day_frame function to get a list of dates, months and years with an index variable of the datetime
    dayframe = day_frame(daylist)
    #day frame contains the references for the netcdf. Using the date range, we create
    #a subset for the range we're interested in
    daylist = dayframe[(dayframe['date'] >= startdate) & (dayframe['date'] <= enddate)].date
    dayframe = dayframe[(dayframe['date'] >= startdate) & (dayframe['date'] <= enddate)]
    #Create a list of unique months
    months = list(dayframe.month.unique())

    ##Method for ordering months and years - extract unique year month
    ##Convert daylist to a list of fortmatted strings
    date_strings = [d.strftime('%Y/%m') for d in daylist]
    #The set conversion gives unique values which are listed in date_months and then sorted
    date_months = list(set(date_strings))
    date_months.sort()
    
    #Calls FILENAME to give a name to output files. But I've commented this out and added a line to the input text
    #name = filename(netc_inputs[0],netc_inputs[-1])
    name = filename
    #MFDataset makes variables spanning multiple files appear as though they were in one file
    nc_open = nc.MFDataset(netc_inputs)
    #Create arrays of latitude and longitude values from the nc4 files
    lat = nc_open.variables['lat'][99:261]
    lon = nc_open.variables['lon'][319:480]
    #lat = nc_open.variables['lat'][:]
    #lon = nc_open.variables['lon'][:]
    
    # Build the output table
    #A table array of zeros with rows of #latx#long (tablelength) and 2 columns
    tab = np.zeros([(len(lat)*len(lon)),2])
    tablength = (len(lat)*len(lon))
    #Set up dates list for later           
    dates = []
    #Iterate through all lat long combinations and populate the table array with all values (tab)   
    count = 0
    for i in range(0,len(lat)):
        for j in range(0,len(lon)):
            tab[count,0] = lat[i]
            tab[count,1] = lon[j]
            count = count + 1
    
    #Iterate through unique list of month-years and summarise climatic variable using GENERAL CALCULATOR.
    for i,monthyear in enumerate(date_months):
        #Identify relevant month & year and print them
        month = np.int64(monthyear.split('/')[1])
        year = np.int64(monthyear.split('/')[0])
        print(str((calendar.month_name[month]+" "+str(year))))
        #Append these individual month-years to "dates" array
        dates.append(str((calendar.month_name[month]+" "+str(year))))
        #Locate relevant cell ranges in dayframe and copy all columns of this to "dateselect" array
        dateselect = dayframe.loc[(dayframe['year'] == year) & (dayframe['month'] == month)]
        #Run general calculator (runoff_component)
        runoff_component = runoff_calc(dateselect,nc_open,clip_domain,tablength)
        #Append output of calculations to "tab" array (first two columns of which are lat/long list)
        tab = np.append(tab,runoff_component,axis=1)
    
    #OPTION USED TO MASK THE OUTPUT - DELETE TO USE ALL THE NETCDF
#    mask = clipmask(clip_domain,lat,lon)
#    tab = tab[mask]
#    tablength = (len(tab))
    #END OF MASK OPTION
    
    #Masked versions of lat/long tables
    mask_lat = np.unique(tab[:,0])
    mask_lon = np.unique(tab[:,1])
    
    #Call the MONTHLY_AVERAGE function to compute monthly averages within the year range and export them to csv
    month_table = monthly_average(tab,months,tablength,name,filepath)
    #Calls MONTHLY_WRITE (and SPATIAL_REFERENCE within that) to create a Geotiff
    monthly_write(month_table,months,mask_lat,mask_lon,name,filepath)
    
    #Creates a csv file of ALL months total rainfall (i.e. not averaged over time period as before, but code is the same)
    outdf = pd.DataFrame(data=tab)
    outdf = outdf.dropna(how='any')
    outdf.columns = [dates[x-2] if x > 1 else x for x in outdf.columns]
    outdf = outdf.rename(columns={0:'Latitude',1:'Longitude'})
    outdf.to_csv(os.path.join(filepath,str(name+"_Monthly_Rain_Data.csv")) ,sep=',',index=False)
    print ("Processing Complete -  Time to Execute  %s:" % (time.time()
    - start_time))
    return ()
    
    
#Call the main function    
if __name__== '__main__':
    main()